# This file is a part of the "Connectors" package
# Copyright (C) 2017-2019 Jonas Schulte-Coerne
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""The connectors package implements decorators for setter and getter methods, so
they can be connected to form a processing chain, in which the result values are
automatically updated, whenever a parameter has been changed. This is similar to
a pipes and filters architecture.
The computations can also be parallelized automatically with threads and processes.
"""

# pylint: disable=wildcard-import;    wildcard imports shall be allowed here, since all submodules have their __all__-variables defined

from ._common import Laziness, Parallelization, executor
from ._decorators import *
from ._helper import *
from ._macro import *

from . import blocks
